
package com.rc.retroweaver;

import jace.parser.*;
import jace.parser.constant.*;
import jace.parser.method.*;
import jace.parser.field.*;
import jace.parser.attribute.*;

import org.apache.bcel.*;
import org.apache.bcel.classfile.*;
import org.apache.bcel.generic.*;
import org.apache.bcel.util.*;

import java.io.*;
import java.util.*;

/**
 * A bytecode enhancer that translates Java 1.5 class files into Java 1.4 class
 * files. The enhancer performs primarily two tasks:
 *
 * 1) Reverses changes made to the class file format in 1.5 to the 
 * former 1.4 format.
 *
 * 2) Replaces compiler generated calls into the new 1.5 runtime with calls
 * into the RetroWeavers replacement runtime.
 *
 * @author Toby Reyelts
 *
 */
public class RetroWeaver {

  private ClassParser parser;
  private JavaClass javaClass;
  private ClassGen generator;
  private String className;
  private ConstantPoolGen constantPool;

  private String sourcePath;
  private String outputPath;

  private static final String newLine = System.getProperty( "line.separator" );
  private static Map/*<Type,String>*/ boxTypes = new HashMap/*<Type,String>*/();

  private static final String AUTOBOX_CLASS = "com.rc.retroweaver.runtime.Autobox";
  private static final String VALUE_OF_METHOD = "valueOf";
  private static final String CLASS_LITERAL_CLASS = "com.rc.retroweaver.runtime.ClassLiteral";
  private static final String GET_CLASS_METHOD = "getClass";
  private static final String GET_CLASS_SIGNATURE = "(Ljava/lang/String;)Ljava/lang/Class;";

  static {
    boxTypes.put( Type.getType( Boolean.class ), "(Z)Ljava/lang/Boolean;" );
    boxTypes.put( Type.getType( Byte.class ), "(B)Ljava/lang/Byte;" );
    boxTypes.put( Type.getType( Character.class ), "(C)Ljava/lang/Character;" );
    boxTypes.put( Type.getType( Short.class ), "(S)Ljava/lang/Short;" );
    boxTypes.put( Type.getType( Integer.class ), "(I)Ljava/lang/Integer;" );
    boxTypes.put( Type.getType( Long.class ), "(J)Ljava/lang/Long;" );
    boxTypes.put( Type.getType( Float.class ), "(F)Ljava/lang/Float;" );
    boxTypes.put( Type.getType( Double.class ), "(D)Ljava/lang/Double;" );
  }

  public void weave( String sourcePath, String outputPath ) throws IOException {

    this.sourcePath = sourcePath;

    if ( outputPath == null ) {
      outputPath = sourcePath;
    }

    this.outputPath = outputPath;
    System.out.println( "[RetroWeaver] Weaving " + sourcePath );

    fixupFormat();

    // System.out.println( "BCEL Parsing file" );

    // We read from the outputPath, because that is where the new fixed up class file lives
    parser = new ClassParser( this.outputPath );
    javaClass = parser.parse();

    generator = new ClassGen( javaClass );
    className = generator.getClassName();
    constantPool = generator.getConstantPool();

    if ( alreadyAspected() ) {
      return;
    }

    // System.out.println( "Fixing up runtime" );
    fixupRuntimeCalls();

    // System.out.println( "Dumping class" );
    generator.getJavaClass().dump( this.outputPath );
  }

  /**
   * Replaces '+' characters with '$' characters in the identifier string.
   * (See http://forum.java.sun.com/thread.jsp?forum=316&thread=499645 for more info on "+"s
   * and identifiers).
   *
   * Replaces references to StringBuilder with StringBuffer - StringBuilder is 1.5 only.
   * I believe the compiler doesn't use StringBuilder in a way that is incompatible with
   * StringBuffer - this has yet to be confirmed by Neal, though. The ultra-safe way of
   * doing this is to create a compatible clone of StringBuilder for 1.4.
   *
   * @param index - a constant pool index to a UTF8Constant.
   */
  private void updateId( jace.parser.ConstantPool pool, int index ) {
    UTF8Constant constant = ( UTF8Constant ) pool.getConstantAt( index );
    String name = constant.toString();
    String newName = name;

    if ( name.indexOf( '+' ) != -1 ) {
      newName = name.replace( '+', '$' );
    }

    if ( newName.indexOf( "java/lang/StringBuilder" ) != -1 ) {
      newName = newName.replace( "java/lang/StringBuilder", "java/lang/StringBuffer" );
    }

    if ( newName != name ) {
      constant.setValue( newName );
      // System.out.println( "replacing " + name + " with " + newName );
    }
  }

  /**
   * Changes the format of the class file from 1.5 to 1.4.
   *
   * 1) Changes the version flag to 1.4
   * 2) Rewrites synthetic access specifiers as synthetic attributes
   *
   * I have to use Jace here, because I don't see how I can make this
   * work with BCEL (easily anyway).
   *
   */
  private void fixupFormat() throws IOException {

    // Read the new class file format spec for how the version is computed.
    final int VERSION_1_4 = 48;

    ClassFile classFile = new ClassFile( sourcePath );
    classFile.setVersion( VERSION_1_4, 0 );

    // Fix up identifiers in the constant pool.

    // Change the class name and class file if necessary
    //
    if ( classFile.getClassName().indexOf( '+' ) != -1 ) {

      String newClassName = classFile.getClassName().replace( '+', '$' );
      classFile.setClassName( newClassName );
      String simpleName = "";

      // System.out.println( "Setting new class name: " + newClassName );

      StringTokenizer st = new StringTokenizer( newClassName, "/" );
      while ( st.hasMoreTokens() ) {
        simpleName = st.nextToken();
      }

      File f = new File( outputPath );

      // Delete the old class file, if we're weaving to the same place.
      if ( sourcePath.equals( outputPath ) ) {
        // System.out.println( "deleting " + sourcePath );
        f.delete();
      }

      File dir = f.getParentFile();
      String name = f.getName();

      outputPath = dir.getCanonicalPath() + File.separator + simpleName + ".class";

      // System.out.println( "Updating output path: " + outputPath );
    }

    // Change the super class name if necessary
    //
    if ( classFile.getSuperClassName().indexOf( '+' ) != -1 ) {
      String newClassName = classFile.getSuperClassName().replace( '+', '$' );
      classFile.setSuperClassName( newClassName );
    }

    // Run through all of the CONSTANT_Class, CONSTANT_Field, and CONSTANT_Method 
    // pool entries, and update their identifers. This catches the interfaces 
    // implemented by the class too.
    //
    jace.parser.ConstantPool pool = classFile.getConstantPool();

    for ( int i = 0; i < pool.getSize(); ++i ) {

      jace.parser.constant.Constant c = pool.getConstant( i );

      if ( c instanceof ClassConstant ) {
        ClassConstant constant = ( ClassConstant ) c;
        updateId( pool, constant.getNameIndex() );
      }
      else if ( c instanceof NameAndTypeConstant ) {
        NameAndTypeConstant constant = ( NameAndTypeConstant ) c;
        updateId( pool, constant.getNameIndex() );
        updateId( pool, constant.getDescriptorIndex() );
      }
      else {
        continue;
      }
    }

    // Now run through all of the methods, fields, and local variables
    // updating their names and signatures.
    //

    for ( ClassMethod m : classFile.getMethods() ) {

      updateId( pool, m.getNameIndex() );
      updateId( pool, m.getDescriptorIndex() );

      CodeAttribute code = m.getCode();

      if ( code != null ) {
        LocalVariableTableAttribute lvt = code.getLocalVariableTable();
        if ( lvt != null ) {
          for ( LocalVariableTableAttribute.Variable v : lvt.getVariables() ) {
            updateId( pool, v.nameIndex() );
            updateId( pool, v.descriptorIndex() );
          }
        }
      }
    }

    for ( ClassField f : classFile.getFields() ) {
      updateId( pool, f.getNameIndex() );
      updateId( pool, f.getDescriptorIndex() );
    }

    classFile.writeClass( outputPath );
  }

  /**
   * Replaces calls to the 1.5 runtime with calls to the Retroweaver runtime.
   * Fixes autoboxing, class literals, (_future_ enums too?)
   */
  private void fixupRuntimeCalls() throws IOException {

    // Fix up code in all of the methods
    //
    Method[] methods = generator.getMethods();
    for ( int i = 0; i < methods.length; ++i ) {
      Method m = methods[ i ];

      MethodGen methodGen = new MethodGen( m, className, constantPool );
      InstructionList list = methodGen.getInstructionList();

      // It's possible that this is a native or abstract method
      // and has no instructions associated with it.
      // In that case, it's not going to be a match.
      if ( list == null ) {
        continue;
      }

      fixupAutoboxing( list );
      fixupClassLiterals( list );

      methodGen.setMaxLocals();
      methodGen.setMaxStack();

      generator.removeMethod( m );
      generator.addMethod( methodGen.getMethod() );
      list.dispose();
    }
  }

  /**
   * Fix autoboxing.
   *
   * Search for calls to invokestatic on any of the <Primitive>.valueOf functions
   * and replace them with calls to our own runtime.
   */
  private void fixupAutoboxing( InstructionList list ) {

    InstructionFinder finder = new InstructionFinder( list );
    InstructionFinder.CodeConstraint constraint = new InstructionFinder.CodeConstraint() {
      public boolean checkCode( InstructionHandle[] match ) {
        INVOKESTATIC instruction = ( INVOKESTATIC ) match[ 0 ].getInstruction();
        String methodName = instruction.getMethodName( constantPool );
        Type t = instruction.getType( constantPool );
        Type[] argTypes = instruction.getArgumentTypes( constantPool );
        // System.out.println( "type = " + t );
        boolean matches = 
          methodName.equals( VALUE_OF_METHOD ) && 
          boxTypes.containsKey( t ) &&
          argTypes.length == 1 &&
          argTypes[ 0 ] instanceof BasicType;

        // System.out.println( "match = " + matches );
        return matches;
      }
    };

    Iterator it = finder.search( "INVOKESTATIC", constraint );

    if ( ! it.hasNext() ) {
      return;
    }

    while ( it.hasNext() ) {
      InstructionHandle[] instructions = ( InstructionHandle[] ) it.next();
      InstructionHandle match = instructions[ 0 ];
      INVOKESTATIC instruction = ( INVOKESTATIC ) match.getInstruction();
      Type t = instruction.getType( constantPool );
      constantPool.addClass( AUTOBOX_CLASS );
      int methodIndex = constantPool.addMethodref( AUTOBOX_CLASS, VALUE_OF_METHOD, ( String ) boxTypes.get( t ) );
      INVOKESTATIC replacementInstruction = new INVOKESTATIC( methodIndex );
      InstructionHandle handle = list.insert( match, replacementInstruction );

      try {
        list.delete( match );
      }
      catch( TargetLostException e ) {
        InstructionHandle[] targets = e.getTargets();
        for ( int i = 0; i < targets.length; i++ ) {
          InstructionTargeter[] targeters = targets[ i ].getTargeters();
   
          for ( int j = 0; j < targeters.length; j++ ) {
            targeters[ j ].updateTarget( targets[ i ], handle );
          }
        }
      }
    }
  }

  /**
   * Fix class literals.
   *
   * The 1.5 VM has had its ldc* instructions updated so that it knows how to deal with 
   * CONSTANT_Class in addition to the other types. So, we have to search for uses of 
   * ldc* that point to a CONSTANT_Class and replace them with calls to our runtime.
   *
   */
  private void fixupClassLiterals( InstructionList list ) {

    final String searchTerm = "(LDC | LDC_W)";

    InstructionFinder finder = new InstructionFinder( list );
    InstructionFinder.CodeConstraint constraint = new InstructionFinder.CodeConstraint() {
      public boolean checkCode( InstructionHandle[] match ) {

        if ( match.length != 1 ) {
          // System.out.println( "Ignoring a match of length: " + match.length );
          return false;
        }

        Instruction instruction = match[ 0 ].getInstruction();

        if ( ! ( instruction instanceof LDC ) ) { // catches both LDC and LDC_W
          return false;
        }

        LDC ldc = ( LDC ) instruction;
        org.apache.bcel.classfile.Constant c = constantPool.getConstant( ldc.getIndex() );

        return c instanceof ConstantClass;
      }
    };

    Iterator it = finder.search( searchTerm, constraint );

    if ( ! it.hasNext() ) {
      return;
    }

    while ( it.hasNext() ) {
      InstructionHandle[] instructions = ( InstructionHandle[] ) it.next();
      InstructionHandle match = instructions[ 0 ];
      LDC instruction = ( LDC ) match.getInstruction();
      ConstantClass c = ( ConstantClass ) constantPool.getConstant( instruction.getIndex() );
      ConstantUtf8 cUtf8 = ( ConstantUtf8 ) constantPool.getConstant( c.getNameIndex() );
      String className = cUtf8.getBytes();
      int classNameIndex = constantPool.addString( className );

      // System.out.println( "Replacing an LDC for CONSTANT_Class " + className );
      constantPool.addClass( CLASS_LITERAL_CLASS );
      int methodIndex = constantPool.addMethodref( CLASS_LITERAL_CLASS, GET_CLASS_METHOD, GET_CLASS_SIGNATURE );

      // TODO: Need to dispose newInstructions, but when is it safe??!!
      InstructionList newInstructions = new InstructionList();
      newInstructions.append( new LDC_W( classNameIndex ) );
      newInstructions.append( new INVOKESTATIC( methodIndex ) );
      InstructionHandle handle = list.insert( match, newInstructions );

      try {
        list.delete( match );
      }
      catch( TargetLostException e ) {
        InstructionHandle[] targets = e.getTargets();
        for ( int i = 0; i < targets.length; i++ ) {
          InstructionTargeter[] targeters = targets[ i ].getTargeters();
   
          for ( int j = 0; j < targeters.length; j++ ) {
            targeters[ j ].updateTarget( targets[ i ], handle );
          }
        }
      }

      finder.reread();
      it = finder.search( searchTerm, constraint );
    }
  }

  private boolean alreadyAspected() {

    // do nothing for now
    // I could add in an attribute to look for here

    return false; 
  }

  public static String getUsage() {

    /*
    StringBuffer buf = new StringBuffer();

    buf.append( "Usage: RetroWeaver " );
    buf.append( newLine );
    buf.append( "  <source path>" );
    buf.append( newLine );
    buf.append( "  <output path>" );

    return buf.toString();
    */

    return "Usage: RetroWeaver " + newLine +
           "  <source path>" + newLine +
           "  [<output path>]";
  }

  public static void main( String[] args ) {

    if ( args.length < 1 ) {
      System.out.println( getUsage() );
      return;
    }

    String sourcePath = args[ 0 ];
    String outputPath = null;

    if ( args.length > 1 ) {
      outputPath = args[ 1 ];
    }

    try {
      RetroWeaver weaver = new RetroWeaver();
      weaver.weave( sourcePath, outputPath );
    }
    catch ( Throwable t ) {
      t.printStackTrace();
    }
  }

}

